(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    112833,       3307]
NotebookOptionsPosition[    108910,       3190]
NotebookOutlinePosition[    109325,       3206]
CellTagsIndexPosition[    109282,       3203]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[TextData[{
 "Reference: Maciej Maliborski and Andrzej Rostworowski, ",
 StyleBox["\[OpenCurlyDoubleQuote]Lecture Notes on Turbulent Instability of \
Anti-de Sitter Spacetime\[CloseCurlyDoubleQuote]",
  FontSlant->"Italic"]
}], "Text",
 CellChangeTimes->{{3.5846803289472437`*^9, 3.5846803851255503`*^9}}],

Cell[CellGroupData[{

Cell[TextData[{
 "Series expansion at ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"\[Pi]", "/", "2"}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "=", "3"}], TraditionalForm]]]
}], "Subsection",
 CellChangeTimes->{{3.580207611474031*^9, 3.580207651040331*^9}, {
   3.580289243708642*^9, 3.580289245701269*^9}, 3.583778799752409*^9, {
   3.583850513102908*^9, 3.583850519876788*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Clear", "[", "eq", "]"}], ";", 
  RowBox[{"Clear", "[", 
   RowBox[{
   "\[Phi]", ",", "\[CapitalPi]", ",", "A", ",", "\[Delta]", ",", 
    "\[CapitalPhi]", ",", "d"}], "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.567438779086945*^9, 3.567438782617177*^9}, 
   3.568441678432835*^9, {3.580108216757914*^9, 3.580108217891696*^9}, {
   3.5801083036359873`*^9, 3.580108303863789*^9}, {3.5802176277987432`*^9, 
   3.580217630261375*^9}}],

Cell["Define number of space dimensions", "Text",
 CellChangeTimes->{{3.5838490326644173`*^9, 3.583849039564921*^9}, {
  3.583850497024042*^9, 3.583850498869409*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"d", "=", "3"}], ";"}]], "Input",
 CellChangeTimes->{{3.580108305130987*^9, 3.5801083071828823`*^9}}],

Cell["an the field equations", "Text",
 CellChangeTimes->{{3.58384902465842*^9, 3.58384902833867*^9}, {
  3.583850500809058*^9, 3.583850505312956*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"eq", "[", "1", "]"}], "=", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{
      SubscriptBox["\[PartialD]", "x"], 
      RowBox[{"A", "[", 
       RowBox[{"t", ",", "x"}], "]"}]}]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"d", "-", "2"}], ")"}], "+", 
       RowBox[{"2", 
        SuperscriptBox[
         RowBox[{"Sin", "[", "x", "]"}], "2"]}]}], 
      RowBox[{
       RowBox[{"Sin", "[", "x", "]"}], " ", 
       RowBox[{"Cos", "[", "x", "]"}]}]], 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"A", "[", 
        RowBox[{"t", ",", "x"}], "]"}]}], ")"}]}], "-", 
    RowBox[{
     RowBox[{"Sin", "[", "x", "]"}], 
     RowBox[{"Cos", "[", "x", "]"}], " ", 
     RowBox[{"A", "[", 
      RowBox[{"t", ",", "x"}], "]"}], 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        RowBox[{"\[CapitalPhi]", "[", 
         RowBox[{"t", ",", "x"}], "]"}], "2"], " ", "+", " ", 
       SuperscriptBox[
        RowBox[{"\[CapitalPi]", "[", 
         RowBox[{"t", ",", "x"}], "]"}], "2"]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"eq", "[", "2", "]"}], "=", 
   RowBox[{
    RowBox[{
     SubscriptBox["\[PartialD]", "t"], 
     RowBox[{"A", "[", 
      RowBox[{"t", ",", "x"}], "]"}]}], "+", 
    RowBox[{"2", 
     RowBox[{"Sin", "[", "x", "]"}], 
     RowBox[{"Cos", "[", "x", "]"}], 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", 
       RowBox[{"\[Delta]", "[", 
        RowBox[{"t", ",", "x"}], "]"}]}]], " ", 
     SuperscriptBox[
      RowBox[{"A", "[", 
       RowBox[{"t", ",", "x"}], "]"}], "2"], "  ", 
     RowBox[{"\[CapitalPhi]", "[", 
      RowBox[{"t", ",", "x"}], "]"}], 
     RowBox[{"\[CapitalPi]", "[", 
      RowBox[{"t", ",", "x"}], "]"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"eq", "[", "3", "]"}], "=", 
   RowBox[{
    RowBox[{
     SubscriptBox["\[PartialD]", "x"], 
     RowBox[{"\[Delta]", "[", 
      RowBox[{"t", ",", "x"}], "]"}]}], "+", 
    RowBox[{
     RowBox[{"Sin", "[", "x", "]"}], 
     RowBox[{"Cos", "[", "x", "]"}], 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        RowBox[{"\[CapitalPhi]", "[", 
         RowBox[{"t", ",", "x"}], "]"}], "2"], " ", "+", " ", 
       SuperscriptBox[
        RowBox[{"\[CapitalPi]", "[", 
         RowBox[{"t", ",", "x"}], "]"}], "2"]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"eq", "[", "4", "]"}], "=", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{
      SubscriptBox["\[PartialD]", "t"], 
      RowBox[{"\[CapitalPi]", "[", 
       RowBox[{"t", ",", "x"}], "]"}]}]}], "+", 
    RowBox[{
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", 
       RowBox[{"\[Delta]", "[", 
        RowBox[{"t", ",", "x"}], "]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{"(", 
           RowBox[{"d", "-", "1"}], ")"}], "-", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", " ", 
             RowBox[{"A", "[", 
              RowBox[{"t", ",", "x"}], "]"}]}], ")"}], 
           RowBox[{"Cos", "[", 
            RowBox[{"2", "x"}], "]"}]}]}], 
         RowBox[{
          RowBox[{"Sin", "[", "x", "]"}], 
          RowBox[{"Cos", "[", "x", "]"}]}]], "  ", 
        RowBox[{"\[CapitalPhi]", "[", 
         RowBox[{"t", ",", "x"}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"A", "[", 
         RowBox[{"t", ",", "x"}], "]"}], 
        RowBox[{
         SubscriptBox["\[PartialD]", "x"], 
         RowBox[{"\[CapitalPhi]", "[", 
          RowBox[{"t", ",", "x"}], "]"}]}]}]}], ")"}]}]}]}], " ", 
  ";"}]}], "Input",
 CellChangeTimes->{{3.5674380817764*^9, 3.5674380874031878`*^9}, {
   3.567438168396984*^9, 3.567438180062791*^9}, {3.567438750285215*^9, 
   3.567438775475184*^9}, {3.5674388306070147`*^9, 3.5674388461247873`*^9}, {
   3.567451387252081*^9, 3.567451387799376*^9}, {3.568439456896097*^9, 
   3.5684394622517023`*^9}, 3.568441808349678*^9, {3.568448999708819*^9, 
   3.568449024847969*^9}, {3.580107755074956*^9, 3.58010775695175*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"eq", "[", "All", "]"}], "=", 
   RowBox[{
    RowBox[{
     RowBox[{"eq", "[", "#", "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", "4", "]"}]}]}], ";"}]], "Input",
 CellChangeTimes->{{3.568440214730154*^9, 3.568440244419181*^9}, {
   3.5684490285336313`*^9, 3.568449029300403*^9}, 3.56845062587103*^9, 
   3.568450686087344*^9, 3.568450841227713*^9, 3.5684515250300817`*^9, {
   3.568549553973393*^9, 3.568549573713304*^9}, 3.568549721517901*^9, {
   3.580107763523958*^9, 3.5801077678068733`*^9}, {3.5807455778203883`*^9, 
   3.5807455813073883`*^9}}],

Cell["Definition of auxiliary fields", "Text",
 CellChangeTimes->{{3.580107879660675*^9, 3.580107893159067*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"\[CapitalPhi]", "[", 
    RowBox[{"t_", ",", "x_"}], "]"}], "=", 
   RowBox[{"D", "[", 
    RowBox[{
     RowBox[{"\[Phi]", "[", 
      RowBox[{"t", ",", "x"}], "]"}], ",", "x"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[CapitalPi]", "[", 
    RowBox[{"t_", ",", "x_"}], "]"}], "=", 
   RowBox[{
    RowBox[{"1", "/", 
     RowBox[{"A", "[", 
      RowBox[{"t", ",", "x"}], "]"}]}], 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"\[Delta]", "[", 
      RowBox[{"t", ",", "x"}], "]"}]], 
    RowBox[{"D", "[", 
     RowBox[{
      RowBox[{"\[Phi]", "[", 
       RowBox[{"t", ",", "x"}], "]"}], ",", "t"}], "]"}]}]}], ";"}]}], "Input",\

 CellChangeTimes->{{3.580108227423864*^9, 3.5801082286866093`*^9}, {
  3.580217754739863*^9, 3.580217755818699*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"TraditionalForm", "/@", 
   RowBox[{"eq", "[", "All", "]"}]}], "//", "CellPrint"}]], "Input",
 CellChangeTimes->{{3.5846808477981443`*^9, 3.584680856956881*^9}}],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"-", 
    RowBox[{
     SuperscriptBox["A", 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "(", 
     RowBox[{"t", ",", "x"}], ")"}]}], "-", 
   RowBox[{
    RowBox[{"sin", "(", "x", ")"}], " ", 
    RowBox[{"cos", "(", "x", ")"}], " ", 
    RowBox[{"A", "(", 
     RowBox[{"t", ",", "x"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      FractionBox[
       RowBox[{
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{"\[Delta]", "(", 
           RowBox[{"t", ",", "x"}], ")"}]}]], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox["\[Phi]", 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "(", 
          RowBox[{"t", ",", "x"}], ")"}], "2"]}], 
       SuperscriptBox[
        RowBox[{"A", "(", 
         RowBox[{"t", ",", "x"}], ")"}], "2"]], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox["\[Phi]", 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "(", 
        RowBox[{"t", ",", "x"}], ")"}], "2"]}], ")"}]}], "+", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       RowBox[{
        SuperscriptBox["sin", "2"], "(", "x", ")"}]}], "+", "1"}], ")"}], " ", 
    RowBox[{"csc", "(", "x", ")"}], " ", 
    RowBox[{"sec", "(", "x", ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"A", "(", 
       RowBox[{"t", ",", "x"}], ")"}]}], ")"}]}]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.584682089400477*^9}],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{
    SuperscriptBox["A", 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "(", 
    RowBox[{"t", ",", "x"}], ")"}], "+", 
   RowBox[{"2", " ", 
    RowBox[{"sin", "(", "x", ")"}], " ", 
    RowBox[{"cos", "(", "x", ")"}], " ", 
    RowBox[{"A", "(", 
     RowBox[{"t", ",", "x"}], ")"}], " ", 
    RowBox[{
     SuperscriptBox["\[Phi]", 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "(", 
     RowBox[{"t", ",", "x"}], ")"}], " ", 
    RowBox[{
     SuperscriptBox["\[Phi]", 
      TagBox[
       RowBox[{"(", 
        RowBox[{"1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "(", 
     RowBox[{"t", ",", "x"}], ")"}]}]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.584682089407515*^9}],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{
    RowBox[{"sin", "(", "x", ")"}], " ", 
    RowBox[{"cos", "(", "x", ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      FractionBox[
       RowBox[{
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{"\[Delta]", "(", 
           RowBox[{"t", ",", "x"}], ")"}]}]], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox["\[Phi]", 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "(", 
          RowBox[{"t", ",", "x"}], ")"}], "2"]}], 
       SuperscriptBox[
        RowBox[{"A", "(", 
         RowBox[{"t", ",", "x"}], ")"}], "2"]], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox["\[Phi]", 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "(", 
        RowBox[{"t", ",", "x"}], ")"}], "2"]}], ")"}]}], "+", 
   RowBox[{
    SuperscriptBox["\[Delta]", 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "(", 
    RowBox[{"t", ",", "x"}], ")"}]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.5846820894112883`*^9}],

Cell[BoxData[
 FormBox[
  RowBox[{
   FractionBox[
    RowBox[{
     RowBox[{
      SuperscriptBox["A", 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "(", 
      RowBox[{"t", ",", "x"}], ")"}], " ", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"\[Delta]", "(", 
       RowBox[{"t", ",", "x"}], ")"}]], " ", 
     RowBox[{
      SuperscriptBox["\[Phi]", 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "(", 
      RowBox[{"t", ",", "x"}], ")"}]}], 
    SuperscriptBox[
     RowBox[{"A", "(", 
      RowBox[{"t", ",", "x"}], ")"}], "2"]], "-", 
   FractionBox[
    RowBox[{
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"\[Delta]", "(", 
       RowBox[{"t", ",", "x"}], ")"}]], " ", 
     RowBox[{
      SuperscriptBox["\[Delta]", 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "(", 
      RowBox[{"t", ",", "x"}], ")"}], " ", 
     RowBox[{
      SuperscriptBox["\[Phi]", 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "(", 
      RowBox[{"t", ",", "x"}], ")"}]}], 
    RowBox[{"A", "(", 
     RowBox[{"t", ",", "x"}], ")"}]], "-", 
   FractionBox[
    RowBox[{
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"\[Delta]", "(", 
       RowBox[{"t", ",", "x"}], ")"}]], " ", 
     RowBox[{
      SuperscriptBox["\[Phi]", 
       TagBox[
        RowBox[{"(", 
         RowBox[{"2", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "(", 
      RowBox[{"t", ",", "x"}], ")"}]}], 
    RowBox[{"A", "(", 
     RowBox[{"t", ",", "x"}], ")"}]], "+", 
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"-", 
      RowBox[{"\[Delta]", "(", 
       RowBox[{"t", ",", "x"}], ")"}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"A", "(", 
        RowBox[{"t", ",", "x"}], ")"}], " ", 
       RowBox[{
        SuperscriptBox["\[Phi]", 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "2"}], ")"}],
          Derivative],
         MultilineFunction->None], "(", 
        RowBox[{"t", ",", "x"}], ")"}]}], "+", 
      RowBox[{
       RowBox[{"csc", "(", "x", ")"}], " ", 
       RowBox[{"sec", "(", "x", ")"}], " ", 
       RowBox[{
        SuperscriptBox["\[Phi]", 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "(", 
        RowBox[{"t", ",", "x"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"2", "-", 
         RowBox[{
          RowBox[{"cos", "(", 
           RowBox[{"2", " ", "x"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{"A", "(", 
             RowBox[{"t", ",", "x"}], ")"}]}], ")"}]}]}], ")"}]}]}], 
     ")"}]}]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.584682089416479*^9}],

Cell["\<\
Definition of the series expansion of metric functions and matter fields\
\>", "Text",
 CellChangeTimes->{{3.5801077853330517`*^9, 3.580107786914035*^9}, {
  3.580107926815784*^9, 3.580107954683177*^9}, {3.583849054087698*^9, 
  3.583849054405459*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"me", "=", "7"}], ";", 
  RowBox[{"(*", 
   RowBox[{"Order", " ", "of", " ", "the", " ", "expansion"}], 
   "*)"}]}]], "Input",
 CellChangeTimes->{{3.5802844404318447`*^9, 3.5802844490374126`*^9}, 
   3.5807457999023886`*^9, 3.580745978349388*^9, 3.5807462420553885`*^9, {
   3.5846808950985003`*^9, 3.584680904327723*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"#", "[", 
       RowBox[{"t_", ",", "x_"}], "]"}], ":=", 
      RowBox[{
       RowBox[{"Sum", "[", 
        RowBox[{
         RowBox[{
          RowBox[{
           SubscriptBox["#", "i"], "[", "t", "]"}], 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             RowBox[{"\[Pi]", "/", "2"}]}], ")"}], "i"]}], ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "0", ",", "me"}], "}"}]}], "]"}], "+", 
       SuperscriptBox[
        RowBox[{"O", "[", 
         RowBox[{"x", ",", 
          RowBox[{"\[Pi]", "/", "2"}]}], "]"}], 
        RowBox[{"me", "+", "1"}]]}]}], ")"}], "&"}], "/@", 
   RowBox[{"{", 
    RowBox[{"A", ",", "\[Delta]", ",", "\[Phi]"}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.58010780188385*^9, 3.580107902053694*^9}, {
  3.580208030047783*^9, 3.580208042123543*^9}, {3.580217845443084*^9, 
  3.580217845564934*^9}, {3.580284452866415*^9, 3.580284456666597*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"TraditionalForm", "[", 
     RowBox[{"(", 
      RowBox[{"#", "\[Rule]", 
       RowBox[{"#", "[", 
        RowBox[{"t", ",", "x"}], "]"}]}], ")"}], "]"}], "&"}], "/@", 
   RowBox[{"{", 
    RowBox[{"A", ",", "\[Delta]", ",", "\[Phi]"}], "}"}]}], "//", 
  "CellPrint"}]], "Input",
 CellChangeTimes->{{3.5802080588223057`*^9, 3.5802080939189053`*^9}, {
  3.584681488592787*^9, 3.5846815251565742`*^9}}],

Cell[BoxData[
 FormBox[
  RowBox[{"A", "\[Rule]", 
   InterpretationBox[
    RowBox[{
     RowBox[{
      SubscriptBox["A", "0"], "(", "t", ")"}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "1"], "(", "t", ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"x", "-", 
        FractionBox["\[Pi]", "2"]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "2"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "2"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "3"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "3"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "4"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "4"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "5"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "5"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "6"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "7"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
     InterpretationBox[
      RowBox[{"O", "(", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"x", "-", 
          FractionBox["\[Pi]", "2"]}], ")"}], "8"], ")"}],
      SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
      Editable->False]}],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {
      Subscript[$CellContext`A, 0][$CellContext`t], 
      Subscript[$CellContext`A, 1][$CellContext`t], 
      Subscript[$CellContext`A, 2][$CellContext`t], 
      Subscript[$CellContext`A, 3][$CellContext`t], 
      Subscript[$CellContext`A, 4][$CellContext`t], 
      Subscript[$CellContext`A, 5][$CellContext`t], 
      Subscript[$CellContext`A, 6][$CellContext`t], 
      Subscript[$CellContext`A, 7][$CellContext`t]}, 0, 8, 1],
    Editable->False]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.584682089674836*^9}],

Cell[BoxData[
 FormBox[
  RowBox[{"\[Delta]", "\[Rule]", 
   InterpretationBox[
    RowBox[{
     RowBox[{
      SubscriptBox["\[Delta]", "0"], "(", "t", ")"}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "1"], "(", "t", ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"x", "-", 
        FractionBox["\[Pi]", "2"]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "2"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "2"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "3"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "3"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "4"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "4"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "5"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "5"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "6"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Delta]", "7"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
     InterpretationBox[
      RowBox[{"O", "(", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"x", "-", 
          FractionBox["\[Pi]", "2"]}], ")"}], "8"], ")"}],
      SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
      Editable->False]}],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {
      Subscript[$CellContext`\[Delta], 0][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 1][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 2][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 3][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 4][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 5][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 6][$CellContext`t], 
      Subscript[$CellContext`\[Delta], 7][$CellContext`t]}, 0, 8, 1],
    Editable->False]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.5846820896787643`*^9}],

Cell[BoxData[
 FormBox[
  RowBox[{"\[Phi]", "\[Rule]", 
   InterpretationBox[
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "(", "t", ")"}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "1"], "(", "t", ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"x", "-", 
        FractionBox["\[Pi]", "2"]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "2"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "2"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "3"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "3"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "4"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "4"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "5"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "5"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "6"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
     RowBox[{
      RowBox[{
       SubscriptBox["\[Phi]", "7"], "(", "t", ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
     InterpretationBox[
      RowBox[{"O", "(", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"x", "-", 
          FractionBox["\[Pi]", "2"]}], ")"}], "8"], ")"}],
      SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
      Editable->False]}],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {
      Subscript[$CellContext`\[Phi], 0][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 1][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 2][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 3][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 4][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 5][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 6][$CellContext`t], 
      Subscript[$CellContext`\[Phi], 7][$CellContext`t]}, 0, 8, 1],
    Editable->False]}], TraditionalForm]],
 GeneratedCell->True,
 CellAutoOverwrite->True,
 CellChangeTimes->{3.5846820896821012`*^9}],

Cell[TextData[{
 "The total mass of the system is defined as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["lim", 
     RowBox[{"x", "\[Rule]", " ", 
      RowBox[{"\[Pi]", "/", "2"}]}]], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "A"}], ")"}], "sin", " ", 
     RowBox[{"x", " ", "/", " ", 
      SuperscriptBox["cos", "3"]}], "x", " "}]}], TraditionalForm]]],
 ". We require the total mass to be finite. Then: "
}], "Text",
 CellChangeTimes->{{3.5802860662892027`*^9, 3.580286096861652*^9}, {
  3.5807923906197877`*^9, 3.5807925801937876`*^9}, {3.580792617060788*^9, 
  3.580792671623788*^9}, {3.583644400805088*^9, 3.5836444691126833`*^9}, {
  3.583644567600444*^9, 3.583644595113069*^9}, {3.583849086915758*^9, 
  3.583849103658461*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    SubscriptBox["A", "0"], "[", "t_", "]"}], ":=", "1"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    SubscriptBox["A", "1"], "[", "t_", "]"}], ":=", "0"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    SubscriptBox["A", "2"], "[", "t_", "]"}], ":=", "0"}], ";"}]}], "Input",
 CellChangeTimes->{{3.5807460698403883`*^9, 3.5807461587233887`*^9}, 
   3.5807462031303883`*^9, {3.583849142344448*^9, 3.583849168262124*^9}}],

Cell[TextData[{
 "and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["A", "3"], "(", "t", ")"}], "=", 
    RowBox[{"-", 
     RowBox[{"M", "(", "t", ")"}]}]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.5802860662892027`*^9, 3.580286096861652*^9}, {
  3.5807923906197877`*^9, 3.5807925801937876`*^9}, {3.580792617060788*^9, 
  3.580792671623788*^9}, {3.583644400805088*^9, 3.5836444691126833`*^9}, {
  3.5836446029306498`*^9, 3.58364460989529*^9}, {3.5837264999197702`*^9, 
  3.583726508849348*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"A", "[", 
  RowBox[{"t", ",", "x"}], "]"}]], "Input",
 CellChangeTimes->{{3.5807462224363885`*^9, 3.5807462263723884`*^9}}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"1", "+", 
   RowBox[{
    RowBox[{
     SubscriptBox["A", "3"], "[", "t", "]"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], ")"}], "3"]}], "+", 
   RowBox[{
    RowBox[{
     SubscriptBox["A", "4"], "[", "t", "]"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], ")"}], "4"]}], "+", 
   RowBox[{
    RowBox[{
     SubscriptBox["A", "5"], "[", "t", "]"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], ")"}], "5"]}], "+", 
   RowBox[{
    RowBox[{
     SubscriptBox["A", "6"], "[", "t", "]"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
   RowBox[{
    RowBox[{
     SubscriptBox["A", "7"], "[", "t", "]"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
   InterpretationBox[
    SuperscriptBox[
     RowBox[{"O", "[", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], "]"}], "8"],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
    Editable->False]}],
  SeriesData[$CellContext`x, Rational[1, 2] Pi, {1, 0, 0, 
    Subscript[$CellContext`A, 3][$CellContext`t], 
    Subscript[$CellContext`A, 4][$CellContext`t], 
    Subscript[$CellContext`A, 5][$CellContext`t], 
    Subscript[$CellContext`A, 6][$CellContext`t], 
    Subscript[$CellContext`A, 7][$CellContext`t]}, 0, 8, 1],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.5807462283533883`*^9, 3.5807463135183887`*^9, 3.580790183456788*^9, 
   3.580791024699788*^9, {3.5836446180650597`*^9, 3.583644627948923*^9}, 
   3.583644679377001*^9, 3.583849171180282*^9, 3.583850529831746*^9, 
   3.584681176495281*^9, 3.58468153455503*^9, 3.584682090684903*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Series", "[", 
  RowBox[{
   RowBox[{"eq", "[", "All", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"\[Pi]", "/", "2"}], ",", 
     RowBox[{"-", "1"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.580284239674829*^9, 3.580284322367587*^9}, {
   3.580284360328765*^9, 3.580284367044256*^9}, {3.580284421377597*^9, 
   3.5802844298275137`*^9}, {3.580284550936988*^9, 3.580284553173402*^9}, {
   3.5802845886938543`*^9, 3.5802846021486397`*^9}, 3.580284649568368*^9, {
   3.5802859668378067`*^9, 3.580285972734539*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    SuperscriptBox[
     RowBox[{"O", "[", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], "]"}], "1"],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 1, 1, 1],
    Editable->False], ",", 
   InterpretationBox[
    SuperscriptBox[
     RowBox[{"O", "[", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], "]"}], "1"],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 1, 1, 1],
    Editable->False], ",", 
   InterpretationBox[
    SuperscriptBox[
     RowBox[{"O", "[", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], "]"}], "0"],
    SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 0, 1],
    Editable->False], ",", 
   InterpretationBox[
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SubscriptBox["\[Phi]", "1"], "[", "t", "]"}]}], ")"}]}], 
       RowBox[{"x", "-", 
        FractionBox["\[Pi]", "2"]}]]}], "+", 
     InterpretationBox[
      SuperscriptBox[
       RowBox[{"O", "[", 
        RowBox[{"x", "-", 
         FractionBox["\[Pi]", "2"]}], "]"}], "0"],
      SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, -1, 0, 1],
      Editable->False]}],
    SeriesData[$CellContext`x, Rational[1, 2] 
     Pi, {(-2) E^(-Subscript[$CellContext`\[Delta], 0][$CellContext`t]) 
      Subscript[$CellContext`\[Phi], 1][$CellContext`t]}, -1, 0, 1],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.58028424061069*^9, 3.5802843227379007`*^9}, 
   3.5802843676487427`*^9, {3.580284417369935*^9, 3.580284460390397*^9}, 
   3.580284501454862*^9, {3.5802845462872257`*^9, 3.58028455372253*^9}, 
   3.580284602674279*^9, 3.580284650042677*^9, 3.58028473122011*^9, 
   3.5802859732031097`*^9, 3.580286111453331*^9, 3.58028814951973*^9, 
   3.580288603564302*^9, 3.5807463767663884`*^9, 3.580790207736788*^9, 
   3.580791030550788*^9, 3.583644705363706*^9, 3.583849173393009*^9, 
   3.5838505299230947`*^9, 3.584681180462129*^9, 3.5846815362718973`*^9, 
   3.584682091078331*^9}]
}, Open  ]],

Cell[TextData[{
 "For convenience, we implement the function which transforms results \
returned by ",
 StyleBox["Solve", "Input"],
 "  and ",
 StyleBox["SolveAlways", "Input"],
 " into functions."
}], "Text",
 CellChangeTimes->{{3.583849337379751*^9, 3.583849357606736*^9}, {
  3.583849406294471*^9, 3.5838495182372026`*^9}, {3.584680973236843*^9, 
  3.5846809770003*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "ToFunction", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"ToFunction", "[", "expr_Rule", "]"}], ":=", 
    RowBox[{"expr", "/.", 
     RowBox[{
      RowBox[{"Rule", "[", 
       RowBox[{
        RowBox[{"a_", "[", "t_", "]"}], ",", "e_"}], "]"}], "\[RuleDelayed]", 
      
      RowBox[{"Rule", "[", 
       RowBox[{"a", ",", 
        RowBox[{"Function", "[", 
         RowBox[{"t", ",", "e"}], "]"}]}], "]"}]}]}]}], ";"}], 
  "\[IndentingNewLine]", 
  RowBox[{"(*", 
   RowBox[{
   "Turn", " ", "off", " ", "some", " ", "annoying", " ", "warnings"}], 
   "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"SolveAlways", "::", "ifun"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"Solve", "::", "svars"}], "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.580286311183773*^9, 3.58028633002324*^9}, {
   3.580286373038069*^9, 3.5802864683669863`*^9}, {3.5802881575710173`*^9, 
   3.58028816466406*^9}, {3.5802882149841957`*^9, 3.5802882210279284`*^9}, {
   3.58384919774537*^9, 3.5838492161973763`*^9}, 3.583849264608779*^9}],

Cell[TextData[{
 "We store results in as a list of rules in ",
 StyleBox["cnd", "Input"],
 " variable."
}], "Text",
 CellChangeTimes->{{3.583849545107214*^9, 3.583849573157425*^9}, {
  3.5838504750191717`*^9, 3.583850482336845*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580286179704965*^9, 3.580286188110755*^9}, {
  3.5802866788601294`*^9, 3.580286679577984*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "/.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", 
        RowBox[{"-", "1"}]}], "}"}]}], "]"}], "\[Equal]", "0"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.5802861188360567`*^9, 3.580286136355986*^9}, {
  3.580286192687366*^9, 3.580286194075061*^9}, {3.580286492213711*^9, 
  3.580286509576461*^9}, {3.580286694195331*^9, 3.580286702643972*^9}, {
  3.580288607874208*^9, 3.580288613463174*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True", ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"-", 
         RowBox[{
          SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
       RowBox[{
        SubscriptBox["\[Phi]", "1"], "[", "t", "]"}]}], 
      RowBox[{
       RowBox[{"-", 
        FractionBox["\[Pi]", "2"]}], "+", "x"}]]}], "\[Equal]", "0"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.580286127084086*^9, 3.580286136788093*^9}, 
   3.580286194617424*^9, {3.5802864909743357`*^9, 3.580286509971057*^9}, {
   3.5802867030043373`*^9, 3.5802867250861683`*^9}, 3.5802881688686247`*^9, 
   3.58028822306026*^9, 3.580288614222411*^9, 3.580289417953411*^9, 
   3.580791079001788*^9, 3.580791370953788*^9, 3.580791920159788*^9, 
   3.583644731848674*^9, 3.583849576608939*^9, 3.583850530088482*^9, 
   3.584681191684062*^9, 3.58468154048636*^9, 3.584682091883574*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SubscriptBox["\[Phi]", "1"], "\[Rule]", 
   RowBox[{"Function", "[", 
    RowBox[{"t", ",", "0"}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{{3.580286127084086*^9, 3.580286136788093*^9}, 
   3.580286194617424*^9, {3.5802864909743357`*^9, 3.580286509971057*^9}, {
   3.5802867030043373`*^9, 3.5802867250861683`*^9}, 3.5802881688686247`*^9, 
   3.58028822306026*^9, 3.580288614222411*^9, 3.580289417953411*^9, 
   3.580791079001788*^9, 3.580791370953788*^9, 3.580791920159788*^9, 
   3.583644731848674*^9, 3.583849576608939*^9, 3.583850530088482*^9, 
   3.584681191684062*^9, 3.58468154048636*^9, 3.5846820918867292`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["\[Phi]", "1"], "\[Rule]", 
     RowBox[{"Function", "[", 
      RowBox[{"t", ",", "0"}], "]"}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580286197737631*^9, 3.5802862095484657`*^9}, {
  3.580286279994363*^9, 3.580286285933384*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "/.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "0"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.580286216058878*^9, 3.5802862165098257`*^9}, {
   3.580286517392663*^9, 3.5802865642090607`*^9}, {3.580286734266358*^9, 
   3.580286736913659*^9}, 3.580288620771044*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", 
   RowBox[{
    RowBox[{
     SubscriptBox["\[Delta]", "1"], "[", "t", "]"}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"-", 
        RowBox[{
         SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
      RowBox[{
       SubscriptBox["\[Phi]", "2"], "[", "t", "]"}]}], "-", 
     RowBox[{
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{
        SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]], " ", 
      RowBox[{
       SuperscriptBox[
        SubscriptBox["\[Delta]", "0"], "\[Prime]",
        MultilineFunction->None], "[", "t", "]"}], " ", 
      RowBox[{
       SuperscriptBox[
        SubscriptBox["\[Phi]", "0"], "\[Prime]",
        MultilineFunction->None], "[", "t", "]"}]}], "-", 
     RowBox[{
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{
        SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]], " ", 
      RowBox[{
       SuperscriptBox[
        SubscriptBox["\[Phi]", "0"], "\[Prime]\[Prime]",
        MultilineFunction->None], "[", "t", "]"}]}]}], "\[Equal]", "0"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.580286213673979*^9, 3.580286217420177*^9}, 
   3.580286287463367*^9, {3.580286518443055*^9, 3.580286564803343*^9}, {
   3.580286737491123*^9, 3.580286749790966*^9}, 3.580287907744137*^9, {
   3.580288200977676*^9, 3.580288224465239*^9}, 3.580288621669375*^9, 
   3.580289420637806*^9, 3.5807919460497875`*^9, 3.5836447381773367`*^9, 
   3.583849586125093*^9, 3.5838505305645037`*^9, 3.584681198513268*^9, 
   3.5846815445487423`*^9, 3.584682092189042*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["\[Delta]", "1"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "2"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{
       RowBox[{"-", 
        FractionBox["1", "2"]}], " ", 
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"2", " ", 
         RowBox[{
          SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "0"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Phi]", "0"], "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "t", "]"}]}], ")"}]}]}], "]"}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.580286213673979*^9, 3.580286217420177*^9}, 
   3.580286287463367*^9, {3.580286518443055*^9, 3.580286564803343*^9}, {
   3.580286737491123*^9, 3.580286749790966*^9}, 3.580287907744137*^9, {
   3.580288200977676*^9, 3.580288224465239*^9}, 3.580288621669375*^9, 
   3.580289420637806*^9, 3.5807919460497875`*^9, 3.5836447381773367`*^9, 
   3.583849586125093*^9, 3.5838505305645037`*^9, 3.584681198513268*^9, 
   3.5846815445487423`*^9, 3.584682092192853*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         RowBox[{"-", 
          FractionBox["1", "2"]}], " ", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"2", " ", 
           RowBox[{
            SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox[
             SubscriptBox["\[Phi]", "0"], "\[Prime]\[Prime]",
             MultilineFunction->None], "[", "t", "]"}]}], ")"}]}]}], 
       "]"}]}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580286570970443*^9, 3.580286572616826*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "//.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "1"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.58028676531249*^9, 3.580286767299444*^9}, 
   3.580286869152719*^9, 3.5802886380355864`*^9, {3.580792037037788*^9, 
   3.580792040003788*^9}, {3.584681555852084*^9, 3.584681556533432*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"2", " ", 
       RowBox[{
        SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{
       SuperscriptBox[
        SubscriptBox["\[Phi]", "0"], "\[Prime]",
        MultilineFunction->None], "[", "t", "]"}], "2"]}], "\[Equal]", "0"}], 
   ",", "True", ",", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        RowBox[{
         SubscriptBox["\[Delta]", "2"], "[", "t", "]"}]}], "-", 
       RowBox[{
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Phi]", "0"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}], "2"]}]}], ")"}]}], 
    "\[Equal]", "0"}], ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{
  3.5802865953213778`*^9, {3.580286768204474*^9, 3.580286775661099*^9}, 
   3.580286869777265*^9, 3.580287911082384*^9, 3.580288233248733*^9, 
   3.580288639057287*^9, 3.5802894238294907`*^9, 3.5807919900007877`*^9, 
   3.5807920439557877`*^9, 3.583644744291499*^9, 3.5838495919694033`*^9, 
   3.583850530788665*^9, 3.584681203421489*^9, 3.58468155712039*^9, 
   3.58468209290991*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["\[Delta]", "0"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{"-", "\[Infinity]"}]}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "2"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.5802865953213778`*^9, {3.580286768204474*^9, 3.580286775661099*^9}, 
   3.580286869777265*^9, 3.580287911082384*^9, 3.580288233248733*^9, 
   3.580288639057287*^9, 3.5802894238294907`*^9, 3.5807919900007877`*^9, 
   3.5807920439557877`*^9, 3.583644744291499*^9, 3.5838495919694033`*^9, 
   3.583850530788665*^9, 3.584681203421489*^9, 3.58468155712039*^9, 
   3.5846820930491467`*^9}]
}, Open  ]],

Cell[TextData[{
 "Our (half)automatic solver solves ",
 Cell[BoxData[
  RowBox[{
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"2", " ", 
      RowBox[{
       SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", 
       FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      SuperscriptBox[
       SubscriptBox["\[Phi]", "0"], "\[Prime]",
       MultilineFunction->None], "[", "t", "]"}], "2"]}], "\[Equal]", "0"}]], 
  "Output",
  CellChangeTimes->{
   3.5802865953213778`*^9, {3.580286768204474*^9, 3.580286775661099*^9}, 
    3.580286869777265*^9, 3.580287911082384*^9, 3.580288233248733*^9, 
    3.580288639057287*^9, 3.5802894238294907`*^9, 3.5807919900007877`*^9, 
    3.5807920439557877`*^9, 3.583644744291499*^9}],
 " giving ",
 Cell[BoxData[
  RowBox[{
   SubscriptBox["\[Delta]", "0"], "\[Rule]", 
   RowBox[{"Function", "[", 
    RowBox[{"t", ",", 
     RowBox[{"-", "\[Infinity]"}]}], "]"}]}]], "Output",
  CellChangeTimes->{
   3.5802865953213778`*^9, {3.580286768204474*^9, 3.580286775661099*^9}, 
    3.580286869777265*^9, 3.580287911082384*^9, 3.580288233248733*^9, 
    3.580288639057287*^9, 3.5802894238294907`*^9, 3.5807919900007877`*^9, 
    3.5807920439557877`*^9, 3.58364474429397*^9}],
 " and that is not what we want here. Thus at this point we have to put ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "'"}], 
     RowBox[{"(", "t", ")"}]}], "=", "0"}], TraditionalForm]]],
 " condition maually:"
}], "Text",
 CellChangeTimes->{{3.5802886646359587`*^9, 3.580288711221239*^9}, 
   3.580289433118161*^9, {3.583644783116612*^9, 3.583644846535665*^9}, {
   3.583644887827619*^9, 3.5836450487896214`*^9}, {3.58384959980337*^9, 
   3.583849601320883*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        SubscriptBox["f", "0"]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580286804585679*^9, 3.580286825068445*^9}, {
  3.5802882793330793`*^9, 3.580288279634542*^9}, {3.583645111857476*^9, 
  3.583645127772148*^9}, {3.584681302543648*^9, 3.584681304010412*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "//.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "2"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.58028688220082*^9, 3.580286902259396*^9}, {
  3.580288719520258*^9, 3.580288748676001*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", 
   RowBox[{
    RowBox[{"3", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "2"], " ", 
     RowBox[{
      SubscriptBox["\[Delta]", "3"], "[", "t", "]"}]}], "\[Equal]", "0"}], 
   ",", 
   RowBox[{
    RowBox[{"4", " ", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", 
       RowBox[{
        SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "2"], " ", 
     RowBox[{
      SubscriptBox["\[Phi]", "4"], "[", "t", "]"}]}], "\[Equal]", "0"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.580286830638124*^9, 3.580286843308955*^9}, {
   3.580286882831667*^9, 3.58028689252673*^9}, 3.580287912923253*^9, {
   3.580288285565557*^9, 3.580288295977294*^9}, {3.580288720433096*^9, 
   3.58028874989732*^9}, 3.58028943548875*^9, 3.580792071543788*^9, 
   3.583645131953249*^9, 3.583849842893062*^9, 3.5838505309745283`*^9, 
   3.5846812155095673`*^9, 3.584681306767741*^9, 3.584681680357697*^9, 
   3.584682093719844*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["\[Delta]", "3"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.580286830638124*^9, 3.580286843308955*^9}, {
   3.580286882831667*^9, 3.58028689252673*^9}, 3.580287912923253*^9, {
   3.580288285565557*^9, 3.580288295977294*^9}, {3.580288720433096*^9, 
   3.58028874989732*^9}, 3.58028943548875*^9, 3.580792071543788*^9, 
   3.583645131953249*^9, 3.583849842893062*^9, 3.5838505309745283`*^9, 
   3.5846812155095673`*^9, 3.584681306767741*^9, 3.584681680357697*^9, 
   3.584682093722177*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        SubscriptBox["f", "0"]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580286915791436*^9, 3.580286916023435*^9}, 
   3.580288305090341*^9, 3.58468131185081*^9}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "//.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "3"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{
  3.580286922948475*^9, {3.580288326701284*^9, 3.580288334558957*^9}, {
   3.580288785842012*^9, 3.580288787379142*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "3"]}], " ", 
     RowBox[{
      SubscriptBox["A", "4"], "[", "t", "]"}]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "3"], " ", 
     RowBox[{
      SuperscriptBox[
       SubscriptBox["A", "3"], "\[Prime]",
       MultilineFunction->None], "[", "t", "]"}]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"4", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "3"], " ", 
     RowBox[{
      SubscriptBox["\[Delta]", "4"], "[", "t", "]"}]}], "\[Equal]", "0"}], 
   ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"-", 
        RowBox[{
         SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]]}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "3"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"4", " ", 
        RowBox[{
         SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "-", 
       RowBox[{"10", " ", 
        RowBox[{
         SubscriptBox["\[Phi]", "5"], "[", "t", "]"}]}], "+", 
       RowBox[{
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}], "\[Equal]", 
    "0"}]}], "}"}]], "Output",
 CellChangeTimes->{{3.5802869201606092`*^9, 3.580286923480009*^9}, 
   3.5802879145590067`*^9, {3.5802883071974373`*^9, 3.58028833525863*^9}, 
   3.580288788108633*^9, 3.580289438034186*^9, 3.580792082425788*^9, 
   3.5836451379979277`*^9, 3.583849858677951*^9, 3.583850531162169*^9, 
   3.584681314104011*^9, 3.5846816885881777`*^9, 3.584682093939643*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["A", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SuperscriptBox[
     SubscriptBox["A", "3"], "\[Prime]",
     MultilineFunction->None], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "3"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{
       FractionBox["1", "4"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"10", " ", 
          RowBox[{
           SubscriptBox["\[Phi]", "5"], "[", "t", "]"}]}], "-", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"2", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "-", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"2", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], 
     "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.5802869201606092`*^9, 3.580286923480009*^9}, 
   3.5802879145590067`*^9, {3.5802883071974373`*^9, 3.58028833525863*^9}, 
   3.580288788108633*^9, 3.580289438034186*^9, 3.580792082425788*^9, 
   3.5836451379979277`*^9, 3.583849858677951*^9, 3.583850531162169*^9, 
   3.584681314104011*^9, 3.5846816885881777`*^9, 3.584682094064073*^9}]
}, Open  ]],

Cell[TextData[{
 "In this way we get the condition ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["A", "3"], "'"}], 
     RowBox[{"(", "t", ")"}]}], "=", "0"}], TraditionalForm]]],
 " which implies that total mass is conserved. That is if the mass is finite \
it has to be conserved.  Additionally we obtain that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "5"], "(", "t", ")"}], TraditionalForm]]],
 " is expressed in terms of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "3"], "(", "t", ")"}], TraditionalForm]]],
 " and its derivatives."
}], "Text",
 CellChangeTimes->{{3.580288804650399*^9, 3.580288890792547*^9}, {
  3.580288979751614*^9, 3.580289066488811*^9}, {3.5807921206347876`*^9, 
  3.580792193518788*^9}, {3.583645157986087*^9, 3.583645163812336*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{"0", "==", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         RowBox[{
          SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "-", 
        RowBox[{"10", " ", 
         RowBox[{
          SubscriptBox["\[Phi]", "5"], "[", "t", "]"}]}], "+", 
        RowBox[{
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"2", " ", 
           RowBox[{
            SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Delta]", "0"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Phi]", "3"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}]}], "+", 
        RowBox[{
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"2", " ", 
           RowBox[{
            SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "5"], "[", "t", "]"}]}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.580287102027981*^9, 3.580287142464181*^9}, {
  3.5838499202130013`*^9, 3.583849927590313*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SubscriptBox["\[Phi]", "5"], "\[Rule]", 
   RowBox[{"Function", "[", 
    RowBox[{"t", ",", 
     RowBox[{
      FractionBox["1", "10"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         RowBox[{
          SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "+", 
        RowBox[{
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"2", " ", 
           RowBox[{
            SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Delta]", "0"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Phi]", "3"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}]}], "+", 
        RowBox[{
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"2", " ", 
           RowBox[{
            SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], "]"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.5802871152396393`*^9, 3.580287142992399*^9}, 
   3.580287915301675*^9, 3.580288340607584*^9, 3.580288854023674*^9, 
   3.58028907224846*^9, 3.580289440707759*^9, 3.5807930559437876`*^9, 
   3.58364521599165*^9, {3.5838499233031054`*^9, 3.5838499281239243`*^9}, 
   3.5838505314434757`*^9, 3.5846813396905727`*^9, 3.5846817051906443`*^9, 
   3.5846820943743467`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        SubscriptBox["f", "0"]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{"-", "M"}]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["1", "10"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], 
       "]"}]}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.58028699239748*^9, 3.580286992632782*^9}, 
   3.580287149318293*^9, {3.580288352711347*^9, 3.580288355822015*^9}, {
   3.58028909430904*^9, 3.580289124542642*^9}, 3.580792981774788*^9, 
   3.584681345454514*^9}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "//.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "4"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{
  3.580287164487672*^9, 3.580288449936096*^9, {3.580289133928604*^9, 
   3.5802891352088118`*^9}, {3.580289462349139*^9, 3.580289486383788*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "3"]}], " ", 
      RowBox[{
       SubscriptBox["A", "4"], "[", "t", "]"}]}], "-", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "4"], " ", 
      RowBox[{
       SubscriptBox["A", "5"], "[", "t", "]"}]}]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "4"], " ", 
     RowBox[{
      SuperscriptBox[
       SubscriptBox["A", "4"], "\[Prime]",
       MultilineFunction->None], "[", "t", "]"}]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"4", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "3"], " ", 
      RowBox[{
       SubscriptBox["\[Delta]", "4"], "[", "t", "]"}]}], "+", 
     RowBox[{"5", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "4"], " ", 
      RowBox[{
       SubscriptBox["\[Delta]", "5"], "[", "t", "]"}]}]}], "\[Equal]", "0"}], 
   ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "9"}], " ", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", 
       RowBox[{
        SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "4"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"M", " ", 
        RowBox[{
         SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         SubscriptBox["\[Phi]", "6"], "[", "t", "]"}]}]}], ")"}]}], 
    "\[Equal]", "0"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.580287165119141*^9, 3.580288450853314*^9, {3.580289128920974*^9, 
   3.5802891357751217`*^9}, {3.580289451906904*^9, 3.580289487001791*^9}, 
   3.5807930658557878`*^9, 3.583645219821662*^9, 3.5838499384745398`*^9, 
   3.583850531641724*^9, 3.584681348706633*^9, 3.584681723526004*^9, 
   3.5846820948995457`*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["\[Phi]", "6"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{
       FractionBox["1", "2"], " ", "M", " ", 
       RowBox[{
        SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}]}], "]"}]}], ",", 
   RowBox[{
    SuperscriptBox[
     SubscriptBox["A", "4"], "\[Prime]",
     MultilineFunction->None], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["A", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["A", "5"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "5"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.580287165119141*^9, 3.580288450853314*^9, {3.580289128920974*^9, 
   3.5802891357751217`*^9}, {3.580289451906904*^9, 3.580289487001791*^9}, 
   3.5807930658557878`*^9, 3.583645219821662*^9, 3.5838499384745398`*^9, 
   3.583850531641724*^9, 3.584681348706633*^9, 3.584681723526004*^9, 
   3.584682095013393*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        SubscriptBox["f", "0"]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{"-", "M"}]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["1", "10"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], 
       "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "6"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["1", "2"], " ", "M", " ", 
         RowBox[{
          SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580287184232071*^9, 3.580287184527264*^9}, {
   3.580288456828085*^9, 3.5802884693656187`*^9}, {3.580289144061768*^9, 
   3.580289150027602*^9}, {3.580289493542459*^9, 3.5802894941457*^9}, {
   3.580793084187788*^9, 3.5807930903557878`*^9}, 3.584681354859096*^9}],

Cell[TextData[{
 "Mind that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "6"], "(", "t", ")"}], TraditionalForm]]],
 " (and also higher order even terms of the expansion of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", 
    RowBox[{"(", 
     RowBox[{"t", ",", "x"}], ")"}]}], TraditionalForm]]],
 " at the boundary ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"\[Pi]", "/", "2"}]}], TraditionalForm]]],
 ") do vanish only in the limit ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "\[Rule]", "0"}], TraditionalForm]]],
 ". Thus the boundary behavior of the eigen modes propagating on a fixed AdS \
background (being odd functions with respect to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"\[Pi]", "/", "2"}]}], TraditionalForm]]],
 ") is incompatible with the self-gravitating case. This is true for all odd \
spatial dimension ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", " ", "\[GreaterEqual]", " ", "3"}], TraditionalForm]]],
 ". This means that the basis of the eigen modes of the wave operator on a \
fixed AdS is well adapted to the self-gravitating case ONLY(!) in even number \
of spatial dimensions ",
 Cell[BoxData[
  FormBox["d", TraditionalForm]]],
 ". "
}], "Text",
 CellChangeTimes->{{3.583726593153145*^9, 3.5837271081341333`*^9}, {
  3.583727145613991*^9, 3.5837272602486362`*^9}, {3.5837273040054626`*^9, 
  3.5837274337913322`*^9}, {3.583849956845799*^9, 3.583850034284049*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "//.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "5"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"ToFunction", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"SolveAlways", "[", 
    RowBox[{
     RowBox[{"%", "//.", "cnd"}], ",", "x"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{
  3.5802871916394444`*^9, {3.580288474613411*^9, 3.580288478544365*^9}, {
   3.580288512346154*^9, 3.580288550672086*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "5"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        RowBox[{
         SubscriptBox["A", "6"], "[", "t", "]"}]}], "+", 
       RowBox[{"9", " ", 
        SuperscriptBox[
         RowBox[{
          SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], "2"]}]}], ")"}]}], 
    "\[Equal]", "0"}], ",", "True", ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "5"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"6", " ", 
        RowBox[{
         SubscriptBox["\[Delta]", "6"], "[", "t", "]"}]}], "-", 
       RowBox[{"9", " ", 
        SuperscriptBox[
         RowBox[{
          SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], "2"]}]}], ")"}]}], 
    "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox["1", "30"]}], " ", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", 
       RowBox[{
        SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["\[Pi]", "2"]}], "+", "x"}], ")"}], "5"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"136", " ", 
        RowBox[{
         SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "-", 
       RowBox[{"840", " ", 
        RowBox[{
         SubscriptBox["\[Phi]", "7"], "[", "t", "]"}]}], "+", 
       RowBox[{"32", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"18", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Delta]", "0"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}], "3"], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"21", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"32", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"2", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"33", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox[
           SubscriptBox["\[Delta]", "0"], "\[Prime]",
           MultilineFunction->None], "[", "t", "]"}], "2"], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"12", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"3", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], 
          TagBox[
           RowBox[{"(", "3", ")"}],
           Derivative],
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"18", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Delta]", "0"], "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], 
          TagBox[
           RowBox[{"(", "3", ")"}],
           Derivative],
          MultilineFunction->None], "[", "t", "]"}]}], "+", 
       RowBox[{"3", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{"4", " ", 
          RowBox[{
           SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
        RowBox[{
         SuperscriptBox[
          SubscriptBox["\[Phi]", "3"], 
          TagBox[
           RowBox[{"(", "4", ")"}],
           Derivative],
          MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}], "\[Equal]", 
    "0"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.580287192022729*^9, 3.5802884792936087`*^9, {3.5802885151715107`*^9, 
   3.580288551503849*^9}, 3.580289155859804*^9, 3.5802894968525867`*^9, {
   3.5807931662457876`*^9, 3.580793180768788*^9}, 3.5838500429510202`*^9, 
   3.583850531849206*^9, 3.584681359159595*^9, 3.584682095298502*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["A", "6"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{"3", " ", 
       SuperscriptBox[
        RowBox[{
         SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], "2"]}]}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "6"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{
       FractionBox["3", "2"], " ", 
       SuperscriptBox[
        RowBox[{
         SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], "2"]}]}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "7"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{
       FractionBox["1", "840"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"136", " ", 
          RowBox[{
           SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "+", 
         RowBox[{"32", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"2", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"18", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          SuperscriptBox[
           RowBox[{
            SuperscriptBox[
             SubscriptBox["\[Delta]", "0"], "\[Prime]",
             MultilineFunction->None], "[", "t", "]"}], "3"], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"21", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"32", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"2", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"33", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          SuperscriptBox[
           RowBox[{
            SuperscriptBox[
             SubscriptBox["\[Delta]", "0"], "\[Prime]",
             MultilineFunction->None], "[", "t", "]"}], "2"], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"12", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], 
            TagBox[
             RowBox[{"(", "3", ")"}],
             Derivative],
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"18", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Delta]", "0"], "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], 
            TagBox[
             RowBox[{"(", "3", ")"}],
             Derivative],
            MultilineFunction->None], "[", "t", "]"}]}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
          RowBox[{
           SuperscriptBox[
            SubscriptBox["\[Phi]", "3"], 
            TagBox[
             RowBox[{"(", "4", ")"}],
             Derivative],
            MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], 
     "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.580287192022729*^9, 3.5802884792936087`*^9, {3.5802885151715107`*^9, 
   3.580288551503849*^9}, 3.580289155859804*^9, 3.5802894968525867`*^9, {
   3.5807931662457876`*^9, 3.580793180768788*^9}, 3.5838500429510202`*^9, 
   3.583850531849206*^9, 3.584681359159595*^9, 3.584682095377009*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cnd", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Phi]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "1"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "0"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        SubscriptBox["f", "0"]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "2"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "3"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{"-", "M"}]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["1", "10"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"4", " ", 
            RowBox[{
             SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], 
       "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "6"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["1", "2"], " ", "M", " ", 
         RowBox[{
          SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}]}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "4"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "5"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      SubscriptBox["A", "6"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{"3", " ", 
         SuperscriptBox[
          RowBox[{
           SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], "2"]}]}], "]"}]}], 
     ",", 
     RowBox[{
      SubscriptBox["\[Delta]", "6"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["3", "2"], " ", 
         SuperscriptBox[
          RowBox[{
           SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], "2"]}]}], "]"}]}], 
     ",", 
     RowBox[{
      SubscriptBox["\[Phi]", "7"], "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{"t", ",", 
        RowBox[{
         FractionBox["1", "840"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"136", " ", 
            RowBox[{
             SubscriptBox["\[Phi]", "3"], "[", "t", "]"}]}], "+", 
           RowBox[{"32", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"18", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            SuperscriptBox[
             RowBox[{
              SuperscriptBox[
               SubscriptBox["\[Delta]", "0"], "\[Prime]",
               MultilineFunction->None], "[", "t", "]"}], "3"], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"21", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"32", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"2", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"33", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            SuperscriptBox[
             RowBox[{
              SuperscriptBox[
               SubscriptBox["\[Delta]", "0"], "\[Prime]",
               MultilineFunction->None], "[", "t", "]"}], "2"], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"12", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"3", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], 
              TagBox[
               RowBox[{"(", "3", ")"}],
               Derivative],
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"18", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Delta]", "0"], "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], 
              TagBox[
               RowBox[{"(", "3", ")"}],
               Derivative],
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
           RowBox[{"3", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{"4", " ", 
              RowBox[{
               SubscriptBox["\[Delta]", "0"], "[", "t", "]"}]}]], " ", 
            RowBox[{
             SuperscriptBox[
              SubscriptBox["\[Phi]", "3"], 
              TagBox[
               RowBox[{"(", "4", ")"}],
               Derivative],
              MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}]}], 
       "]"}]}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.580287270116867*^9, 3.5802872703781424`*^9}, {
   3.5802884899890327`*^9, 3.580288501571228*^9}, {3.5802891629578037`*^9, 
   3.5802891689835873`*^9}, {3.580289504590149*^9, 3.5802895055324593`*^9}, {
   3.5807932190597878`*^9, 3.5807932253137875`*^9}, 3.584681372773651*^9}],

Cell[TextData[{
 "We can check if equations are fullfield up to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[ScriptCapitalO]", "(", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"x", "-", 
       FractionBox["\[Pi]", "2"]}], ")"}], "5"], ")"}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.580289178276016*^9, 3.580289284748651*^9}, {
  3.580289837516983*^9, 3.580289842122962*^9}, {3.583850066094492*^9, 
  3.583850069465777*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Normal", "@", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"eq", "[", "All", "]"}], "//.", "cnd"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "5"}], "}"}]}], "]"}], "\[Equal]", 
    "0"}], "]"}]}]], "Input",
 CellChangeTimes->{
  3.580287280601881*^9, {3.5802892953153257`*^9, 3.580289297671488*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{{3.5802872782178507`*^9, 3.58028728092953*^9}, 
   3.580289298646164*^9, 3.5802895091588707`*^9, 3.580793238618788*^9, 
   3.583850076487981*^9, 3.583850532112673*^9, 3.584681376172399*^9, 
   3.584682095840281*^9}]
}, Open  ]],

Cell[TextData[{
 "For clarity of the presentation we pick up only conditions involving single \
symbolic or numerical constants, these rules stored in ",
 StyleBox["ncnd", "Input"],
 " variable."
}], "Text",
 CellChangeTimes->{{3.580289676480451*^9, 3.580289752845291*^9}, {
  3.583850094287591*^9, 3.583850101808578*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ncnd", "=", 
  RowBox[{
   RowBox[{"Select", "[", 
    RowBox[{"cnd", ",", 
     RowBox[{
      RowBox[{"MatchQ", "[", 
       RowBox[{"#", ",", 
        RowBox[{"Rule", "[", 
         RowBox[{"__", ",", 
          RowBox[{"_", "[", 
           RowBox[{"_", ",", 
            RowBox[{"_", "?", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"NumericQ", "[", "#", "]"}], "||", 
                RowBox[{"#", "===", 
                 RowBox[{"-", "M"}]}], "||", 
                RowBox[{"#", "===", 
                 SubscriptBox["f", "0"]}]}], "&"}], ")"}]}]}], "]"}]}], 
         "]"}]}], "]"}], "&"}]}], "]"}], "//", "Sort"}]}]], "Input",
 CellChangeTimes->{{3.580287634870985*^9, 3.5802876482309837`*^9}, {
  3.580289362780599*^9, 3.580289382951036*^9}, {3.580289519108691*^9, 
  3.580289521270076*^9}, {3.580289552496182*^9, 3.580289645477545*^9}, {
  3.5838501355015993`*^9, 3.583850137419548*^9}, {3.5838504373616467`*^9, 
  3.583850442405945*^9}, {3.58468143306314*^9, 3.58468143746037*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SubscriptBox["A", "3"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      RowBox[{"-", "M"}]}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["A", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["A", "5"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "1"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "2"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "3"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Delta]", "5"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "0"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", 
      SubscriptBox["f", "0"]}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "1"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "2"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}], ",", 
   RowBox[{
    SubscriptBox["\[Phi]", "4"], "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{"t", ",", "0"}], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.580289592009947*^9, {3.5802896230477467`*^9, 3.580289646001177*^9}, 
   3.580289731676478*^9, 3.583850103651164*^9, 3.583850138018407*^9, 
   3.5838504430156612`*^9, 3.58385053224868*^9, 3.584681379201274*^9, 
   3.584681438201211*^9, 3.584682095960786*^9}]
}, Open  ]],

Cell["Substitute conditions into series expressions", "Text",
 CellChangeTimes->{{3.5802897682981577`*^9, 3.58028979458766*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"#", "\[Rule]", 
       RowBox[{"#", "[", 
        RowBox[{"t", ",", "x"}], "]"}]}], "/.", "ncnd"}], ")"}], "&"}], "/@", 
   
   RowBox[{"{", 
    RowBox[{"A", ",", "\[Delta]", ",", "\[Phi]"}], "}"}]}], "//", 
  "TableForm"}]], "Input",
 CellChangeTimes->{
  3.580287656598435*^9, {3.580289654565485*^9, 3.580289656784816*^9}}],

Cell[BoxData[
 TagBox[
  TagBox[GridBox[{
     {
      RowBox[{"A", "\[Rule]", 
       InterpretationBox[
        RowBox[{"1", "-", 
         RowBox[{"M", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "3"]}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["A", "6"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["A", "7"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
         InterpretationBox[
          SuperscriptBox[
           RowBox[{"O", "[", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], "]"}], "8"],
          SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
          Editable->False]}],
        SeriesData[$CellContext`x, Rational[1, 2] Pi, {
         1, 0, 0, -$CellContext`M, 0, 0, 
          Subscript[$CellContext`A, 6][$CellContext`t], 
          Subscript[$CellContext`A, 7][$CellContext`t]}, 0, 8, 1],
        Editable->False]}]},
     {
      RowBox[{"\[Delta]", "\[Rule]", 
       InterpretationBox[
        RowBox[{
         RowBox[{
          SubscriptBox["\[Delta]", "0"], "[", "t", "]"}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["\[Delta]", "6"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["\[Delta]", "7"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
         InterpretationBox[
          SuperscriptBox[
           RowBox[{"O", "[", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], "]"}], "8"],
          SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
          Editable->False]}],
        SeriesData[$CellContext`x, Rational[1, 2] Pi, {
          Subscript[$CellContext`\[Delta], 0][$CellContext`t], 0, 0, 0, 0, 0, 
          
          Subscript[$CellContext`\[Delta], 6][$CellContext`t], 
          Subscript[$CellContext`\[Delta], 7][$CellContext`t]}, 0, 8, 1],
        Editable->False]}]},
     {
      RowBox[{"\[Phi]", "\[Rule]", 
       InterpretationBox[
        RowBox[{
         SubscriptBox["f", "0"], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["\[Phi]", "3"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "3"]}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["\[Phi]", "5"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "5"]}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["\[Phi]", "6"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "6"]}], "+", 
         RowBox[{
          RowBox[{
           SubscriptBox["\[Phi]", "7"], "[", "t", "]"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], ")"}], "7"]}], "+", 
         InterpretationBox[
          SuperscriptBox[
           RowBox[{"O", "[", 
            RowBox[{"x", "-", 
             FractionBox["\[Pi]", "2"]}], "]"}], "8"],
          SeriesData[$CellContext`x, Rational[1, 2] Pi, {}, 0, 8, 1],
          Editable->False]}],
        SeriesData[$CellContext`x, Rational[1, 2] Pi, {
          Subscript[$CellContext`f, 0], 0, 0, 
          Subscript[$CellContext`\[Phi], 3][$CellContext`t], 0, 
          Subscript[$CellContext`\[Phi], 5][$CellContext`t], 
          Subscript[$CellContext`\[Phi], 6][$CellContext`t], 
          Subscript[$CellContext`\[Phi], 7][$CellContext`t]}, 0, 8, 1],
        Editable->False]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.5599999999999999]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Column],
  Function[BoxForm`e$, 
   TableForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.580287658985764*^9, 3.5802893063434553`*^9, {3.580289626028075*^9, 
   3.580289657254121*^9}, 3.580289797361846*^9, 3.5838501114249*^9, 
   3.583850445303514*^9, 3.5838505323747396`*^9, 3.584681381017354*^9, 
   3.584681439966832*^9, 3.584682096090218*^9}]
}, Open  ]],

Cell[TextData[{
 "There is no freedom to prescribe boundary condition at the boundary ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"\[Pi]", "/", "2"}]}], TraditionalForm]]],
 " and the series expansion at ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Rule]", 
    RowBox[{"\[Pi]", "/", "2"}]}], TraditionalForm]]],
 " is uniquelly determied by the ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "0"], TraditionalForm]],
  FormatType->"TraditionalForm"],
 ", ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]],
  FormatType->"TraditionalForm"],
 " parameters and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "3"], "(", "t", ")"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Delta]", "0"], "(", "t", ")"}], TraditionalForm]]],
 Cell[BoxData[
  FormBox[Cell[""], TraditionalForm]], "Input"],
 " functions. "
}], "Text",
 CellChangeTimes->{{3.5802898570607767`*^9, 3.58029003879825*^9}, {
  3.583850233628813*^9, 3.583850264691136*^9}, {3.584681399970025*^9, 
  3.584681414137907*^9}}]
}, Open  ]]
},
WindowSize->{740, 867},
WindowMargins->{{Automatic, 128}, {28, Automatic}},
PrivateNotebookOptions->{"VersionedStylesheet"->{"Default.nb"[8.] -> False}},
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[557, 20, 309, 6, 51, "Text"],
Cell[CellGroupData[{
Cell[891, 30, 432, 13, 44, "Subsection"],
Cell[1326, 45, 469, 10, 31, "Input"],
Cell[1798, 57, 166, 2, 30, "Text"],
Cell[1967, 61, 135, 3, 31, "Input"],
Cell[2105, 66, 151, 2, 30, "Text"],
Cell[2259, 70, 4166, 127, 201, "Input"],
Cell[6428, 199, 599, 13, 31, "Input"],
Cell[7030, 214, 112, 1, 30, "Text"],
Cell[7145, 217, 844, 27, 52, "Input"],
Cell[7992, 246, 196, 4, 31, "Input"],
Cell[8191, 252, 1807, 60, 50, InheritFromParent],
Cell[10001, 314, 960, 34, 24, InheritFromParent],
Cell[10964, 350, 1351, 45, 50, InheritFromParent],
Cell[12318, 397, 3122, 106, 74, InheritFromParent],
Cell[15443, 505, 262, 5, 30, "Text"],
Cell[15708, 512, 357, 8, 31, "Input"],
Cell[16068, 522, 1029, 29, 36, "Input"],
Cell[17100, 553, 461, 13, 31, "Input"],
Cell[17564, 568, 2497, 75, 77, InheritFromParent],
Cell[20064, 645, 2618, 75, 77, InheritFromParent],
Cell[22685, 722, 2584, 75, 77, InheritFromParent],
Cell[25272, 799, 767, 19, 57, "Text"],
Cell[26042, 820, 520, 16, 72, "Input"],
Cell[26565, 838, 540, 15, 32, "Text"],
Cell[CellGroupData[{
Cell[27130, 857, 147, 3, 31, "Input"],
Cell[27280, 862, 1933, 56, 80, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[29250, 923, 566, 12, 31, "Input"],
Cell[29819, 937, 2265, 58, 57, "Output"]
}, Open  ]],
Cell[32099, 998, 373, 10, 53, "Text"],
Cell[32475, 1010, 1178, 31, 112, "Input"],
Cell[33656, 1043, 232, 6, 32, "Text"],
Cell[33891, 1051, 205, 5, 31, "Input"],
Cell[CellGroupData[{
Cell[34121, 1060, 815, 21, 52, "Input"],
Cell[34939, 1083, 999, 23, 57, "Output"],
Cell[35941, 1108, 666, 12, 31, "Output"]
}, Open  ]],
Cell[36622, 1123, 336, 9, 31, "Input"],
Cell[CellGroupData[{
Cell[36983, 1136, 723, 19, 52, "Input"],
Cell[37709, 1157, 1690, 43, 34, "Output"],
Cell[39402, 1202, 1574, 40, 47, "Output"]
}, Open  ]],
Cell[40991, 1245, 1412, 40, 78, "Input"],
Cell[CellGroupData[{
Cell[42428, 1289, 743, 19, 52, "Input"],
Cell[43174, 1310, 1649, 47, 44, "Output"],
Cell[44826, 1359, 761, 18, 31, "Output"]
}, Open  ]],
Cell[45602, 1380, 1844, 49, 57, "Text"],
Cell[47449, 1431, 1042, 28, 52, "Input"],
Cell[CellGroupData[{
Cell[48516, 1463, 646, 18, 52, "Input"],
Cell[49165, 1483, 1230, 33, 45, "Output"],
Cell[50398, 1518, 768, 17, 31, "Output"]
}, Open  ]],
Cell[51181, 1538, 1225, 35, 72, "Input"],
Cell[CellGroupData[{
Cell[52431, 1577, 673, 19, 52, "Input"],
Cell[53107, 1598, 2747, 81, 80, "Output"],
Cell[55857, 1681, 2065, 55, 70, "Output"]
}, Open  ]],
Cell[57937, 1739, 851, 24, 74, "Text"],
Cell[CellGroupData[{
Cell[58813, 1767, 1444, 39, 78, "Input"],
Cell[60260, 1808, 1595, 41, 47, "Output"]
}, Open  ]],
Cell[61870, 1852, 2776, 76, 118, "Input"],
Cell[CellGroupData[{
Cell[64671, 1932, 697, 19, 52, "Input"],
Cell[65371, 1953, 2464, 78, 115, "Output"],
Cell[67838, 2033, 1362, 38, 73, "Output"]
}, Open  ]],
Cell[69215, 2074, 3659, 100, 185, "Input"],
Cell[72877, 2176, 1462, 41, 131, "Text"],
Cell[CellGroupData[{
Cell[74364, 2221, 675, 19, 52, "Input"],
Cell[75042, 2242, 6618, 188, 149, "Output"],
Cell[81663, 2432, 6083, 160, 150, "Output"]
}, Open  ]],
Cell[87761, 2595, 9706, 255, 371, "Input"],
Cell[97470, 2852, 445, 12, 43, "Text"],
Cell[CellGroupData[{
Cell[97940, 2868, 442, 13, 31, "Input"],
Cell[98385, 2883, 340, 6, 31, "Output"]
}, Open  ]],
Cell[98740, 2892, 322, 7, 51, "Text"],
Cell[CellGroupData[{
Cell[99087, 2903, 1072, 26, 92, "Input"],
Cell[100162, 2931, 1962, 57, 92, "Output"]
}, Open  ]],
Cell[102139, 2991, 128, 1, 30, "Text"],
Cell[CellGroupData[{
Cell[102292, 2996, 420, 14, 31, "Input"],
Cell[102715, 3012, 5095, 136, 104, "Output"]
}, Open  ]],
Cell[107825, 3151, 1069, 36, 53, "Text"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
